"""
Tools for handling data files, formats, etc.
"""

import gzip
import io
import os
import random

from .misc_utils import remove_suffix, check_true


def gunzip_content(content):
    """ Helper that will gunzip content """
    f_in = io.BytesIO()
    f_in.write(content)
    f_in.seek(0)
    with gzip.GzipFile(fileobj=f_in, mode='rb') as f:
        gunzipped_content = f.read()
    return gunzipped_content.decode('utf-8')


def generate_sample_fastq_content(num=10, length=10):
    """ Creates (pseudo)randomly generated fastq content."""

    content = ''
    bases = 'ACTG'

    for i in range(num):
        content += '@SEQUENCE{} length={}\n'.format(i, length)
        content += ''.join(random.choice(bases) for _ in range(length)) + '\n'
        content += '+\n'
        content += 'I' * length + '\n'

    return content


FASTQ_SUFFIXES = [".fastq", ".fq"]


def generate_sample_fastq_file(filename, num=10, length=10, compressed=None):
    """
    Creates a new fastq file with the given name, containing (pseudo)randomly generated content.

    Example usage:

        generate_sample_fastq_file('fastq_sample.fastq.gz', num=25, length=50)
           creates a new fastq file with 25 sequences, each of length 50.

        generate_sample_fastq_file('fastq_sample.fastq.gz')
           creates a new fastq file with default characteristics (10 sequences, each of length 10).

    Args:
        filename str: the name of a file to create
        num int: the number of random sequences (default 10)
        length int: the length of the random sequences (default 10)

    Returns:
        the filename

    """
    filename, compressed = normalize_suffixes(filename, FASTQ_SUFFIXES, compressed=compressed)
    check_true(isinstance(compressed, bool), "compressed must be one of True, False, or None (for autoselection)")
    content = generate_sample_fastq_content(num=num, length=length)
    if compressed:
        with gzip.open(filename, 'w') as zipfile:
            zipfile.write(content.encode('ascii'))
    else:
        with io.open(filename, 'w') as outfile:
            outfile.write(content)
    return filename


def normalize_suffixes(filename, suffixes, compressed=None):
    """
    Assures that filename has one of the suffix endings, and maybe also a compression ending.

    The file is assumed to need compression if compression is requested or is implied by the given filename.
    If one of the suffixes is not present (possibly followed by a compression suffix),
    the first of the suffixes is added.

    See unit tests for examples.
    """
    if filename.endswith('.gz'):
        check_true(compressed is not False,
                   "compressed cannot be False if .gz is given explicitly.",
                   error_class=RuntimeError)
        compressed = True
        filename = remove_suffix(".gz", filename)
    elif compressed is None:
        compressed = False
    _, ext = os.path.splitext(filename)
    if ext not in suffixes:
        filename = filename + suffixes[0]
    if compressed:
        filename += ".gz"
    return filename, compressed



import argparse

from dcicutils.command_utils import script_catch_errors, ScriptFailure
from .contribution_utils import Contributions, PROJECT_HOME


EPILOG = __doc__


def show_contributors(repo, exclude_fork=None, verbose=False, save_contributors=False, test=False):
    contributions = Contributions(repo=repo, exclude_fork=exclude_fork, verbose=verbose)
    if save_contributors:
        contributions.save_contributor_data()
    contributions.show_repo_contributors(error_class=ScriptFailure if test else None)


def show_contributors_main(*, simulated_args=None):
    parser = argparse.ArgumentParser(  # noqa - PyCharm wrongly thinks the formatter_class is specified wrong here.
        description=(f"Show authors of a specified repository, which will be presumed"
                     f" to have been cloned as a subdirectory of $PROJECT_HOME ({PROJECT_HOME})"),
        epilog=EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument('repo', default=None,
                        help="name of repository to show contributors for")
    parser.add_argument('--exclude', '-x', default=None,
                        help="name of repository that repo was forked from, whose contributors to exclude")
    parser.add_argument('--save-contributors', '-s', action="store_true", default=False,
                        help="whether to store contributor data to CONTRIBUTORS.json")
    parser.add_argument('--test', '-t', action="store_true", default=False,
                        help="whether to treat this as a test, erring if a cache update is needed")
    parser.add_argument('--verbose', '-v', action="store_true", default=False,
                        help="whether to do verbose output while working")
    args = parser.parse_args(args=simulated_args)

    with script_catch_errors():

        show_contributors(repo=args.repo, exclude_fork=args.exclude, verbose=args.verbose,
                          save_contributors=args.save_contributors, test=args.test)



import re
from typing import Optional
from dcicutils.structured_data import Portal


def ref_lookup_strategy(portal: Portal, type_name: str, schema: dict, value: str) -> (int, Optional[str]):
    #
    # FYI: Note this situation WRT object lookups ...
    #
    # /{submitted_id}                # NOT FOUND
    # /UnalignedReads/{submitted_id} # OK
    # /SubmittedFile/{submitted_id}  # OK
    # /File/{submitted_id}           # NOT FOUND
    #
    # /{accession}                   # OK
    # /UnalignedReads/{accession}    # NOT FOUND
    # /SubmittedFile/{accession}     # NOT FOUND
    # /File/{accession}              # OK
    #
    def ref_validator(schema: Optional[dict],
                      property_name: Optional[str], property_value: Optional[str]) -> Optional[bool]:
        """
        Returns False iff the type represented by the given schema, can NOT be referenced by
        the given property name with the given property value, otherwise returns None.

        For example, if the schema is for the UnalignedReads type and the property name
        is accession, then we will return False iff the given property value is NOT a properly
        formatted accession ID. Otherwise, we will return None, which indicates that the
        caller (in dcicutils.structured_data.Portal.ref_exists) will continue executing
        its default behavior, which is to check other ways in which the given type can NOT
        be referenced by the given value, i.e. it checks other identifying properties for
        the type and makes sure any patterns (e.g. for submitted_id or uuid) are ahered to.

        The goal (in structured_data) being to detect if a type is being referenced in such
        a way that cannot possibly be allowed, i.e. because none of its identifying types
        are in the required form (if indeed there any requirements). Note that it is guaranteed
        that the given property name is indeed an identifying property for the given type.
        """
        if property_format := schema.get("properties", {}).get(property_name, {}).get("format"):
            if (property_format == "accession") and (property_name == "accession"):
                if not _is_accession_id(property_value):
                    return False
        return None

    DEFAULT_RESPONSE = (Portal.LOOKUP_DEFAULT, ref_validator)

    if not value:
        return DEFAULT_RESPONSE
    if not schema:
        if not isinstance(portal, Portal) or not (schema := portal.get_schema(type_name)):
            return DEFAULT_RESPONSE
    if schema_properties := schema.get("properties"):
        if schema_properties.get("accession") and _is_accession_id(value):
            # Case: lookup by accession (only by root).
            return Portal.LOOKUP_ROOT, ref_validator
        elif schema_property_info_submitted_id := schema_properties.get("submitted_id"):
            if schema_property_pattern_submitted_id := schema_property_info_submitted_id.get("pattern"):
                if re.match(schema_property_pattern_submitted_id, value):
                    # Case: lookup by submitted_id (only by specified type).
                    return Portal.LOOKUP_SPECIFIED_TYPE, ref_validator
    return DEFAULT_RESPONSE


# This is here for now because of problems with circular dependencies.
# See: smaht-portal/.../schema_formats.py
def _is_accession_id(value: str) -> bool:
    return isinstance(value, str) and re.match(r"^SMA[1-9A-Z]{9}$", value) is not None
